#!/usr/bin/perl -w

$| = 1;

# ddr2nfi.pl   ("DDRescue to NFI")
#
# This is a Perl program which reads a GNU ddrescue 1.x "log file" (really, a
# sector list file), and either the output of "fdisk -lu" or a SpinRite 6.0
# log file (or both).  If it is given a SpinRite log file, then it uses the
# SpinRite data to adjust the ddrescue bad sector list according to which
# sectors SpinRite was (and was not) able to recover.  It also uses the
# SpinRite data to convert physical (LBA) sector numbers to logical NTFS
# sector numbers (relative to the beginning of the partition).
#
# The output is a .BAT file of "nfi" commands.  (NFI is the Microsoft "NTFS
# File Sector Information Utility," which, given the sector number, can tell
# you what file or directory contains that file.)
#
# If no SpinRite log file is given, or if you only used SpinRite on part of
# the drive, then you should also give this program the "fdisk -lu" output for
# this drive (either the original or the recovered copy), so that it can make
# the physical-to-logical sector number conversions for the sectors that are
# not mentioned in the SpinRite log file.
#
# If you give it netierh a SpinRite log file nor the "fdisk -lu" output, then
# it generates nfi commands using the physical sector numbers (leaving it up
# to nfi to do the conversion).
#
# If the physical-to-logical translation is done, then the generated nfi
# commands assume that the NTFS partition is "C:".  If that is not correct,
# you'll need to edit the generated .bat file to correct the drive letter.
# If the physical-to-logical translation cannot be done, then the synatax
# of the generated nfi commands assumes that the disk drive is the first
# disk drive in the computer.
#
# (You may be wondering: what about other file systems?  Well, if you know of
# a similar tool for another file system, please tell me!)
#
#
# Here's how you can use this tool:
#
# FIRST, run ddrescue 1.1 under Linux to recover as much data as possible.
# If it rescues everything successfully, then you are done.
#
# You can tell ddrescue to do multiple passes to maximize the recovered data.
# This is an example of how I do it on my system.  In this example, the sick
# disk drive is /dev/hdg and the recovered image file and the ddrescue
# log file both reside in the current directory:
#
#    # Ensure that DMA is enabled (probably not necessary with recent Linux):
#    hdparm -d1 /dev/hdg
#    # Recover the easy parts first ("-n" skips subdividing error regions):
#    ddrescue -B -n /dev/hdg drive.ima drive.log
#    # Fill in the gaps (with no "-n" specified):
#    ddrescue -B /dev/hdg drive.ima drive.log
#    # I don't know whether these settings help or not; see 'man hdparm':
#    hdparm -a0 -A0 -d1 -k1 -K1 /dev/hdg
#    # You don't really need to save the log file at this point, but I like
#    # to have it for comparison purposes, so I can see what progress has
#    # been made on subsequent steps:
#    cp -p drive.log drive.log.before-raw
#    # Under Linux (at least some versions), if you don't use a 'raw device'
#    # the Linux disk cache will prevent you from reading good sectors which
#    # are part of the same cache block as bad sectors.  They symptom is that
#    # all the "bad areas" appear to be at least 4096 bytes (8 sectors) long.
#    # By using a 'raw device' we can get the granularity down to a single
#    # sector.  See 'man raw' for how the 'raw' command works.
#    # WARNING: USING THE RAW DEVICE REQUIRES VERSION 1.1 OR LATER of ddrescue:
#    raw /dev/raw/raw9 /dev/hdg
#    ddrescue -B -c 1 -r 1 /dev/raw/raw9 drive.ima drive.log
#    # You could quit at this point, but I like to try again, with a couple
#    # of additional retries, to read the bad sectors:
#    ddrescue -B -c 1 -r 2 /dev/raw/raw9 drive.ima drive.log
#    cp -p drive.log drive.log.after-ddrescue
#    # No more 'raw device':
#    raw /dev/raw/raw9 0 0
#
# SECOND (optional), freeze the drive in your freezer overnight, and try
# ddrescue 1.1 again, perhaps like this:
#
#    hdparm -d1 /dev/hdg
#    hdparm -a0 -A0 -d1 -k1 -K1 /dev/hdg
#    raw /dev/raw/raw9 /dev/hdg
#    ddrescue -B -c 1 -r 1 /dev/raw/raw9 drive.ima drive.log
#    ddrescue -B -c 1 -r 1 /dev/raw/raw9 drive.ima drive.log
#    ddrescue -B -c 1 -r 2 /dev/raw/raw9 drive.ima drive.log
#    cp -p drive.log drive.log.after-freezer
#    raw /dev/raw/raw9 0 0
#
# If it rescues everything successfully, then you are done.  If it rescues
# some, but not all, of the data, then I suggest that you freeze the drive
# a second time, and repeat the above step.
#
# THIRD, boot MS-DOS 7 or FreeDOS and run SpinRite on the drive.  Use
# ddr2sr.pl to build a .bat file of SpinRite commands for just the bad areas
# of the drive.
#
# However, if there are lots of bad sectors, you'll want to use a low DynaStat
# setting, to speed up the recovery process.  Simply edit the .bat file
# generated by ddr2sr.pl to add the dynastat setting.  For example, on most
# computers setting dynastat to 5% ("spinrite dynastat 5 ...") will make it
# take about 2 minutes per bad sector instead of 45 minutes per bad sector.
#
# Be sure to save the SpinRite log file!  (Make sure you don't lose it by
# running out of room on the diskette.)  If necessary, you can stop SpinRite
# occasionally and move its log files (SPIN_LOG.*) elsewhere.  If you
# do that, then when SpinRite is done you'll have two or more SPIN_LOG.*
# files saved, so concatinate them together into one big file before using
# this tool.  E.g.:
# copy tmp1\SPIN_LOG.3+tmp2\SPIN_LOG.3+tmp3\SPIN_LOG.3 SPIN_LOG.3
# (Note: "3" is the drive number, as SpinRite numbers disk drives.)
#
# FOURTH, save a copy of the ddrescue log file BEFORE running ddrescue again!
# The SpinRite-recovered sectors were only PARTIALLY recovered, you don't
# want to pretend that they are really good data.  So first, save a copy
# of the ddrescue logfile from before you ran SpinRite.
#
# FIFTH, run ddrescue to add the SpinRite-recovered sectors to the recovered
# disk image file:
#
#    cp -p drive.log log_before_SR
#    hdparm -d1 /dev/hdg
#    hdparm -a0 -A0 -d1 -k1 -K1 /dev/hdg
#    raw /dev/raw/raw9 /dev/hdg
#    ddrescue -B -c 1 -r 3 /dev/raw/raw9 drive.ima drive.log
#    cp -p drive.log log_after_SR
#    raw /dev/raw/raw9 0 0
#
# (Note: Normally, after the above step drive.log probably shows no errors
# at all, because ddrescue should be able to read all the sectors that
# SpinRite repaired.  If not, then it means that the drive is so sick that
# ddrescue couldn't copy some sectors which SpinRite had just fixed, perhaps
# because the drive ran out of spare sectors.)
#
# SIXTH (finally!) you are ready to use this Perl script (srddrnfi.pl).
# It will read the ddrescue logfiles (sector list files) from before and
# after running SpinRite, plus the SpinRite log file, and from them
# determine which sectors were fully recovered by SpinRite, and which
# were not, so that it can generate a .bat file of NFI commands for all the
# unrecovered sectors.
#
# It takes five file names as command line parameters.  The first parameter
# is the name of the output file, which will be a .bat file of nfi commands:
#
#    perl -w srddrnfi.pl nficmds.bat SPIN_LOG.3 fdisk-lu_output log_before_SR log_after_SR
#
# You can omit the 5th parameter if you know that the final ddrescue command
# got no errors (which is the normal case).  If you didn't run SpinRite, then
# specify "-" in place of the 2nd parameter.  You can also specify "-" in
# place of the 3rd parameter, if you are not providing this script with the
# output of "fdisk -lu".
#
#
# This program is UNCOPYRIGHTED.  It was written by:
#
#    Dave Burton
#    Burton Systems Software
#    Tel: +1 919-481-0149
#    Email: dave at burtonsys.com
#
#
# v.1 - 30-Sep-2005 - first version.  -DAB
# v.2 - 30-Oct-2005 - added fdisk-lu_output parameter, and made SPIN_LOG.x optional
#


#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "10, 12-Sep-15";
# the TLIB Version Control version number and last-modified date



$perlver = "3 or earlier";
if ($] =~ /\$\$Revision\:\s*([0-9.]+)\s/) {
   $perlver = $1;  # probably 4.something
} elsif ($] =~ /([0-9][0-9.]*)/) {
   $perlver = $1;  # probably 5.something or 6.something
}


@sv_ARGV = @ARGV;  # save raw parameter list, for display to the user

# display command-line parameters (just once)
$args_shown = 0;
sub showargs {
   if (!$args_shown) {
      print join( " ", "perl $0", @sv_ARGV ) . "\n";  # echo command-line arguments
      $args_shown = 1;
   }
}


# Handle input parameters and open all the files:

# check for '-v' and/or '-p'
$verbose = 0;
$physicalmode = 0;
$clustersize = 8;
$noREMs = 0;
while (($#ARGV >= 0) && ($ARGV[0] =~ /^-[vpc\d]/)) {
   $arg = shift @ARGV;
   if ($arg =~ /v/) {
      $verbose++;
   }
   if ($arg =~ /p/) {
      $physicalmode = 1;
   }
   if ($arg =~ /c/) {
      $noREMs = 1;
   }
   if ($arg =~ /(\d+)/) {
      $x = $1 + 0;
      if (0==$x) {
         $x = 1;
      }
      if ((1==$x) || (2==$x) || (4==$x) || (8==$x) || (16==$x) || (32==$x) || (64==$x) || (128==$x)) {
         $clustersize = $x;
      } else {
         print "ERROR: $x is an illegal cluster size.  Default of 8 used, instead.\n";
      }
   }
}
if ($verbose) {
   &showargs;
   print "You are using Perl version $perlver\n";
}

($nficmds_bat, $spin_log, $fdisk_out, $log_before_SR, $log_after_SR) = @ARGV;
if (!defined $log_after_SR) {
   $log_after_SR= '';
}
if (!defined $log_before_SR) {
   die "ddr2nfi.pl   (\"DDRescue to NFI\")   v.$ver_ordinal_and_date\n" .
       "\n" .
       "Usage:\n" .
       "   perl -w ddr2nfi.pl {options} nficmds.bat SPIN_LOG.x fdisk_out LogBeforeSR LogAfterSR\n" .
       "\n" .
       "You can omit the 5th parameter (LogAfterSR) if you know that the final ddrescue\n" .
       "command got no errors (which is the normal case if SpinRite processed the whole\n" .
       "drive).   If you didn't run SpinRite, then specify \"-\" in place of the 2nd\n" .
       "parameter (SPIN_LOG.x).  The 3rd parameter (fdisk_out) should be the output of\n" .
       "the \"fdisk -lu /dev/...\" command for this drive, or you can specify \"-\" in\n" .
       "place of the 3rd parameter to omit it.\n" .
       "\n" .
       "Note: it often happens that SpinRite successfully recovers a sector but doesn't\n" .
       "mention that sector in its log file.  Presumably, those sectors were read\n" .
       "successfully after retries, without resorting to DynaStat recovery.  Such\n" .
       "sectors contain good data.  However, if you don't run SpinRite on the entire\n" .
       "drive, that will also result in sectors being omitted from the SpinRite log\n" .
       "file, and those sectors do NOT contain good data.  The only way to tell the\n" .
       "difference is from the LogAfterSR file: the sectors containing good data will\n" .
       "have been successfully copied, and the sectors containing bad data will not.\n" .
       "So do not omit the 5th parameter if you didn't run SpinRite on the whole drive.\n" .
       "\n" .
       "{options} may be omitted (the usual case), or \"-p\" to use the physical (lba)\n" .
       "device name and sector number syntax to nfi, or \"-v\" for verbose/debug, or\n" .
       "\"-c\" for concise output (no REM lines), or \"-nn\" where nn is a decimal integer,\n" .
       "to specify the NTFS cluster size in sectors, used to make nficmds.bat more\n" .
       "concise (the default cluster size is 8, for 4096-byte clusters).  To get nfi\n" .
       "commands for every bad sector, use \"-1\".\n";
}
&showargs;
if (('' eq $spin_log) && ('' ne $log_after_SR)) {
   print "ERROR: do not specify LogAfterSR unless you also specify SPIN_LOG.x\n";
   exit(1);
}
if ('-' eq $spin_log) {
   $spin_log = '';
} else {
   open( SPIN_LOG, "<$spin_log") || die "ERROR: Could not read $spin_log, $!\n";
}
if ('-' eq $fdisk_out) {
   $fdisk_out = '';
} else {
   open( FDISK_OUT, "<$fdisk_out") || die "ERROR: Could not read $fdisk_out, $!\n";
}
open( LOG_BEFORE_SR, "<$log_before_SR") || die "ERROR: Could not read $log_before_SR, $!\n";
if ('-' eq $log_after_SR) {
   $log_after_SR = '';
}
if ('' ne $log_after_SR) {
   open( LOG_AFTER_SR, "<$log_after_SR") || die "ERROR: Could not read $log_after_SR, $!\n";
}
open( NFICMDS, ">$nficmds_bat" ) || die "ERROR: Could not create $nficmds_bat, $!\n";
binmode NFICMDS;  # write output in binary mode, so we can force CR+LF line termination regardless of whether we're running under Linux or DOS/Windows



######################################################################
# Here's a sample of what a ddrescue log file looks like:
#
#    # Rescue Logfile. Created by GNU ddrescue version 1.0
#    #      pos        size  status
#    0x00000000  0x0004BE00  +
#    0x0004BE00  0x00000200  -
#    0x0004C000  0x33078E00  +
#    0x330C4E00  0x00001000  -
#    0x330C5E00  0x8ACDF200  +
#
######################################################################


# input is an integer represented as a string (probably hexadecimal preceeded by '0x'),
# result is the integer divided by 512.
sub divby_512 {
   local($dividend) = @_;
   # print "dividend=$dividend\n";
   local($result) = 0;
   if ($dividend =~ /^0x/) {
      # print "tricky way!\n";
      # tricky divide by 512: drop the last 2 hexidecimal digits, and divide by 2 (which is more robust for very large integers)
      $result = substr($dividend,0,length($dividend)-2);
      $result = int(eval($result) / 2);  # gotta use eval() because it starts with 0x
   } else {
      # print "conventional way,\n";
      $result = int((0+$dividend) / 512);
   }
   # print "result=$result\n";
   return $result;
}


# First, read log_after_SR, to obtain the list of sectors which were supposedly
# fixed by SpinRite, but still couldn't be copied by ddrescue.  Comment lines
# and lines ending with anything other than '-' are ignored.
$num_bad_fixed_sectors = 0;
if ('' ne $log_after_SR) {
   if ($verbose) {
      print "reading $log_after_SR ...\n";
   }
   while (<LOG_AFTER_SR>) {
      if (($_ =~ /\d/) && ($_ !~ /^\#/)) {
         $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
         # Parse a ddrescue log file line:
         @tmp = split( /\s+/, $_ );
         # for ($i=0; $i<=$#tmp; $i++) { print "dbg: tmp[$i]='" . $tmp[$i] . "'\n"; }
         ($pos_byt, $len_byt, $status) = @tmp;
         if (($#tmp != 2) || ($pos_byt !~ /\d/) || ($len_byt !~ /\d/)) {
            if ((1 == $#tmp) && ($len_byt =~ /^[\?\*\/\-\+]$/)) {
               # the new status line (introduced in ddrescue 1.6 or 1.7)
               $pos_sect = &divby_512($pos_byt);
               $status = $len_byt;
               print "Status: ";
               if ($status eq '?') {
                  print "copying data";
               } elsif ($status eq '*') {
                  print "trimming bad areas";
               } elsif ($status eq '/') {
                  print "splitting bad areas";
               } elsif ($status eq '-') {
                  print "retrying bad blocks";
               } elsif ($status eq '+') {
                  print "finished";
               } else {
                  print "unrecognized status '$status'";
               }
               print " (status code = '$status'), at sector $pos_sect ('$log_after_SR')\n";
            } else {
               print "ERROR: unrecognized line in '$log_after_SR':\n$_\n";
            }
            # print "ERROR: unrecognized line in '$log_after_SR':\n$_\n";
            next;
         }
         if ('-' ne $tmp[2]) {
            # ignore all but '-' (error) regions
            next;
         }
         if ($len_byt eq '-1') {
            print "ERROR: length=-1 for status = '-' in '$log_after_SR':\n$_";
            next;
         }
         # Now convert byte offsets and counts to LBA sector number & sector count:
         $pos_sect = &divby_512($pos_byt);
         $len_sect = &divby_512($len_byt);
         # print "Uh oh!  '$log_after_SR': Still-bad 'fixed' sect no. = $pos_sect, count = $len_sect, '$status'\n";
         for ($i=0; $i<$len_sect; $i++) {
            $definitely_bad_sectors{$pos_sect} = 1;
            $pos_sect++;
            $num_bad_fixed_sectors++;
         }
      }
   }
   if ('' ne $spin_log) {
      print "$num_bad_fixed_sectors sectors named in '$log_after_SR' should have been fixed by Spinrite, but are still unreadable.\n";
   }
   # if ($num_bad_fixed_sectors) {
   #    print "Note: processing of LogAfterSR is not yet well-tested.\n";
   # }
}
# The result of this step is an associative array called %definitely_bad_sectors,
# and $num_bad_fixed_sectors, which is a count of the bad sectors seen in the
# log_after_SR file.  If SpinRite fixed all the bad sectors, of if log_after_SR
# was omitted, then $num_bad_fixed_sectors==0.



# Next, read log_before_SR, to obtain the list of sectors which were bad before
# SpinRite started its work.  Lines beginning with '#' or ending with anything
# other than '-' are ignored.
$num_originally_bad_sectors = 0;
if ($verbose) {
   print "reading $log_before_SR ...\n";
}
while (<LOG_BEFORE_SR>) {
   if (($_ =~ /\d/) && ($_ !~ /^\#/)) {
      $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
      # Parse a ddrescue log file line:
      @tmp = split( /\s+/, $_ );
      # for ($i=0; $i<=$#tmp; $i++) { print "dbg: tmp[$i]='" . $tmp[$i] . "'\n"; }
      ($pos_byt, $len_byt, $status) = @tmp;
      if (($#tmp != 2) || ($pos_byt !~ /\d/) || ($len_byt !~ /\d/)) {
         if ((1 == $#tmp) && ($len_byt =~ /^[\?\*\/\-\+]$/)) {
            # the new status line (introduced in ddrescue 1.6 or 1.7)
            $pos_sect = &divby_512($pos_byt);
            $status = $len_byt;
            print "Status: ";
            if ($status eq '?') {
               print "copying data";
            } elsif ($status eq '*') {
               print "trimming bad areas";
            } elsif ($status eq '/') {
               print "splitting bad areas";
            } elsif ($status eq '-') {
               print "retrying bad blocks";
            } elsif ($status eq '+') {
               print "finished";
            } else {
               print "unrecognized status '$status'";
            }
            print " (status code = '$status'), at sector $pos_sect ('$log_before_SR')\n";
         } else {
            print "ERROR: unrecognized line in '$log_before_SR':\n$_\n";
         }
         # print "ERROR: unrecognized line in '$log_before_SR':\n$_\n";
         next;
      }
      if ('-' ne $tmp[2]) {
         # ignore all but '-' (error) regions
         next;
      }
      if ($len_byt eq '-1') {
         print "ERROR: length=-1 for status = '-' in '$log_before_SR':\n$_";
         next;
      }
      # Now convert byte offsets and counts to LBA sector number & sector count:
      $pos_sect = &divby_512($pos_byt);
      $len_sect = &divby_512($len_byt);
      # print "dbg: Bad sect no. = $pos_sect, count = $len_sect, '$status'\n";
      for ($i=0; $i<$len_sect; $i++) {
         $originally_bad_sectors{$pos_sect} = 1;
         $num_originally_bad_sectors++;
         $pos_sect++;
      }
   }
}
print "$num_originally_bad_sectors sectors named in '$log_before_SR' ";
if ('' ne $spin_log) {
   print "were bad before Spinrite started its work.\n";
} else {
   print "are unreadable.\n";
}
# The result of this step is an associative array called %originally_bad_sectors,
# and $num_originally_bad_sectors, which is a count of the bad sectors seen in
# log_before_SR file.



# If there was no SpinRite log, then all the originally bad sectors are still bad:
if (('' eq $spin_log) && ('' eq $log_after_SR)) {
   if ($verbose) {
      print "dbg: since there's no spin_log, LogAfterSR = LogBeforeSR\n";
   }
   %definitely_bad_sectors = %originally_bad_sectors;
   $num_bad_fixed_sectors = $num_originally_bad_sectors;
}



# Read FDISK_OUT, to get the partition table information.  Here's an example
# (note the blank line at the beginning):
#----------------------------------------------------------------------
#
# Disk /dev/hdg: 160.0 GB, 160000000000 bytes
# 255 heads, 63 sectors/track, 19452 cylinders, total 312500000 sectors
# Units = sectors of 1 * 512 = 512 bytes
#
#    Device Boot      Start         End      Blocks   Id  System
# /dev/hdg1              63      112454       56196   de  Dell Utility
# /dev/hdg2   *      112455   305154674   152521110    7  HPFS/NTFS
# /dev/hdg3       305154675   312496379     3670852+  db  CP/M / CTOS / ...
#----------------------------------------------------------------------
$saw_Units_line = $num_partitions = $num_driveletters = 0;
if ('' ne $fdisk_out) {
 while (<FDISK_OUT>) {
   $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
   if ($_ =~ /^Units(\:| \=) sectors of 1 \* 512 \= 512 bytes/) {
      $saw_Units_line = 1;
   }
   # match a line like this:
   # '/dev/hdg3       305154675   312496379     3670852+  db  CP/M / CTOS / ...'
   if ($_ =~ /^(\/dev\/\w*[a-z])(\d+)\s+(\*|)\s*(\d+)\s+(\d+)\s+(\d+)(\+|)\s+(\w{1,2})\s*(.*)$/) {
      ($dev, $partnum, $bootind, $start, $end, $blocks, $halfblockind, $id, $system) =
        ($1, $2, $3, $4, $5, $6, $7, $8, $9);
      if ($verbose) {
         print "dbg: device= $dev$partnum\n";
         print "     bootind=$bootind\n";
         print "     start=$start\n";
         print "     end=$end\n";
         print "     blocks=$blocks";
         if ('' ne $halfblockind) {
            print ".5";
         }
         print "\n     id=$id\n";
         print "     system=$system\n";
      }
      if ($system =~ /(NTFS|FAT)/) {
         $driveletter = pack( 'c', $num_driveletters+ord('C') );
         $num_driveletters++;
      } else {
         $driveletter = '?';
      }
      $starting[$num_partitions] = 0+$start;
      $ending[$num_partitions] = 0+$end;
      $driveletters[$num_partitions] = $driveletter;
      $num_partitions++;
   }
 }

 # Result is the partition table, represented by @starting, @ending, and @driveletters
 if (!$saw_Units_line) {
    print "Warning: '$fdisk_out' does not look like the output of \"fdisk -lu\".\n";
 } elsif (!$num_partitions) {
    print "Warning: no recognizable partition table lines seen in '$fdisk_out'.\n";
 }
}


# Using the partition table (in @starting, @ending, and @driveletters),
# look up the partition table index for a given sector number, or -1
# for failure.
sub which_partition {
   local( $sect ) = @_;
   local( $result ) = -1;
   local( $i );
   # print "REM dbg: which_partition($sect), ";
   $result = -1;
   for ($i = 0; $i <= $#starting; $i++) {
      # print "  i=$i ";
      if (($sect >= $starting[$i]) && ($sect <= $ending[$i])) {
         # printf "= result (start=%d <= sect=%d <= end=%d)", $starting[$i], $sect, $ending[$i];
         $result = $i;
         last;
      }
      # printf "(start=%d, end=%d) ", $starting[$i], $ending[$i];
   }
   # print "\n";
   return $result;
}



# look up the drive letter from the sector number
sub driveletter {
   local( $sect ) = @_;
   local( $partnum ) = &which_partition( $sect );
   local( $result ) = '?';
   if ($partnum != -1) {
      $result = $driveletters[$partnum];
   }
   return $result;
}



# Calculate the logical sector number from the physical (LBA) sector number,
# by Using the "fdisk -lu" output.  Return the logcal sector number, or -1
# if unknown.
sub lba_to_logical {
   local( $result ) = -1;
   if ('' ne $fdisk_out) {
      local( $sect ) = @_;
      local( $partnum ) = &which_partition( $sect );
      if ($partnum != -1) {
         $result = $sect - $starting[$partnum];
      }
   }
   return $result;
}



# Read SPIN_LOG, to obtain the list of "partially fixed" sectors.
# These are sectors which are now readable (thanks to SpinRite), but which
# contain damaged or questionable data.  The only lines we care about in
# SPIN_LOG are lines which look like these examples:
#
#    Some data was unrecoverable in this          115,957,327   115,957,264 
#
#    All data recovered from damaged sector.      227,380,454   227,380,391 
#
$count_bad_sectors = $count_good_sectors = 0;
$count_went_bad_after_ddrescue = $count_went_iffy_after_ddrescue = 0;
$tmpcnt = 0;
$prev_partition_offset = 0;
if ('' ne $spin_log) {
 if ($verbose) {
    print "reading $spin_log ...\n";
 }
 while (<SPIN_LOG>) {
   if ($_ =~ /Some data was unrecoverable in this|All data recovered from damaged sector/) {
      $_ =~ s/\s*$//;  # strip trailing lf, whitespace, etc.
      @tmp = split( /\/, $_ );
      if (($tmp[2] !~ /^[0-9\, ]+$/) || ($tmp[3] !~ /^[0-9\, ]+$/)) {
         print "ERROR: unrecognized line in '$spin_log':\n$_";
         next;
      }

      $sector = $tmp[2];
      $sector =~ s/[ \,]*//g;
      $sector += 0;  # convert to integer

      $lsect = $tmp[3];
      $lsect =~ s/[ \,]*//g;
      $lsect += 0;  # convert to integer

      $partition_offset = $sector - $lsect;
      if ($partition_offset != $prev_partition_offset) {
         if ($prev_partition_offset > 0) {
            print "Partition sector offset changed from $prev_partition_offset to $partition_offset in '$spin_log'.\n";
         } else {
            print "Partition sector offset is $partition_offset in '$spin_log'.\n";
         }
         if ($partition_offset < $prev_partition_offset) {
            print "ERROR: Partition offset shrunk!\n";
         }
         if ($verbose) {
            print "Line $. in '$spin_log' is:\n$_\n";
         }
         $prev_partition_offset = $partition_offset;
      }

      # print 'dbg: bad sector #' . $sector . "\n";
      if ($_ =~ /unrecoverable/) {
         if (defined $originally_bad_sectors{$sector}) {
            if (!defined $bad_sectors{$sector}) {
               $bad_sectors{$sector} = $lsect;
               $count_bad_sectors++;
            }
         } else {
            $count_went_bad_after_ddrescue++;
         }
      } else {
         if (defined $originally_bad_sectors{$sector}) {
            if (!defined $good_sectors{$sector}) {
               $good_sectors{$sector} = $lsect;
               $count_good_sectors++;
            }
         } else {
            $count_went_iffy_after_ddrescue++;
         }
      }
      $tmpcnt++;
   }
 }
}
# The results of this step are the following global variables:
#
# %good_sectors        =  keys are the fully recovered physical sector numbers, values
#                         are the logical sector numbers for the fully recovered sectors.
#
# %bad_sectors         =  keys are the unrecoverable (or incompletely recovered) physical
#                         sector numbers, values are the logical sector numbers.
#
# $count_good_sectors  =  $#good_sectors + 1
#
# $count_bad_sectors   =  $#bad_sectors + 1
#
# $count_went_bad_after_ddrescue  =  number of sectors that SpinRite called unrecoverable,
#                         yet had already been successfully copied in the first ddrescue run.
#
# $count_went_iffy_after_ddrescue  =  number of sectors that SpinRite "fully recovered,"
#                         but had already been successfully copied in the first ddrescue run.


if ('' ne $spin_log) {
   if ($tmpcnt > ($count_bad_sectors+$count_good_sectors)) {
      printf "Note: %d sector numbers were duplicates in '$spin_log'\n", $tmpcnt-($count_bad_sectors+$count_good_sectors);
   }
   print "$count_bad_sectors bad ('partially recovered') sector numbers found in '$spin_log'\n";
   print "$count_good_sectors good ('fully recovered') sector numbers found in '$spin_log'\n";
   if ($count_went_bad_after_ddrescue) {
      print "Plus, another $count_went_bad_after_ddrescue sectors were partially unrecoverable by Spinrite but had already been copied by ddrescue, so they didn't matter anyhow.\n";
   }
   if ($count_went_iffy_after_ddrescue) {
      print "Plus, another $count_went_iffy_after_ddrescue sectors were fully recovered by Spinrite but had already been copied by ddrescue, so they didn't matter anyhow.\n";
   }

   if (($count_bad_sectors + $count_good_sectors) < $num_originally_bad_sectors) {
      print "The other " . ($num_originally_bad_sectors-($count_bad_sectors+$count_good_sectors)) . " sectors are presumed to have read correctly after retries.\n";
   }
}


$good_in_log1_but_bad_in_log2 = 0;
foreach $sect (keys %definitely_bad_sectors) {
   if (!defined $originally_bad_sectors{$sect}) {
      $good_in_log1_but_bad_in_log2++;
      die "ERROR: Inconsistent ddrescue log files.  Sector $sect is bad in '$log_after_SR' but good in '$log_before_SR', which should be impossible.\n";
   }
}

$fixed_originally_bad = 0;
foreach $sect (keys %good_sectors) {
   if (defined $bad_sectors{$sect}) {
      print "Warning: sector $sect was listed as both partially recovered and fully recovered in '$spin_log'; we assume the worst.\n";
   } elsif ((defined $originally_bad_sectors{$sect}) && !(defined $definitely_bad_sectors{$sect})) {
      delete $originally_bad_sectors{$sect};
      $fixed_originally_bad++;
   } # else it was already read successfully by ddrescue before SpinRite started, so ignore it
}

# $went_bad_after_ddrescue = 0;
# foreach $sect (keys %bad_sectors) {
#    if (!defined $originally_bad_sectors{$sect}) {
#       # it was already read successfully by ddrescue before SpinRite started, so ignore it
#       delete $bad_sectors{$sect};
#       $went_bad_after_ddrescue++;
#    }
# }
# This is now handled in the loop that reads spin_log -- it won't insert %bad_sectors entries for sectors that aren't in keys %originally_bad_sectors.


$still_bad_after_SR = 0;
$went_bad_again = 0;
foreach $sect (keys %definitely_bad_sectors) {
   $still_bad_after_SR++;
   if (!defined $bad_sectors{$sect}) {
      $went_bad_again++;
      $bad_sectors{$sect} = -1
   }
}
if (('' ne $spin_log) && $went_bad_again) {
   print "Warning: $went_bad_again sectors went bad again between when Spinrite fixed them and\n" .
         "when ddrescue tried to copy them.  This drive is very sick!\n";
}
if ($verbose || ($spin_log ne '') || ($log_after_SR ne '')) {
   print "$still_bad_after_SR sectors are still bad in final ddrescue log file.\n";
}




sub numerically { $a <=> $b }


@final_bad_sector_list = sort numerically keys %bad_sectors;
if ($verbose || ($spin_log ne '') || ($log_after_SR ne '')) {
   printf "%d bad or questionable sectors after all is done.\n", 1+$#final_bad_sector_list;
}

$count_actual_num_nfi_cmds = 0;
$prev_lsect = -1;
$already_warned1 = 0;
foreach $sect (@final_bad_sector_list) {
   $lsect = $bad_sectors{$sect};  # if $lsect > 0, then it was from SpinRite output

   # Deduce the logical sector number from the fdisk -lu output.
   $lsect2 = &lba_to_logical( $sect );
   $drvltr = &driveletter( $sect );

   $lsectstr = '';
   if ($lsect == $lsect2) {
      if ((-1 == $lsect) && (-1 == $lsect2)) {
         $lsectstr = "?";
      } elsif ($verbose) {
         $lsectstr = "$lsect=$lsect2";
      } else {
         $lsectstr = "$lsect";
      }
   } elsif (-1 == $lsect) {
      if ($verbose) {
         $lsectstr = "$lsect2 [fdisk]";
      } else {
         $lsectstr = "$lsect2";
      }
      $lsect = $lsect2;
   } elsif (-1 == $lsect2) {
      if ($verbose) {
         $lsectstr = "$lsect [SR]";
      } else {
         $lsectstr = "$lsect";
      }
   } else {
      # Trouble!  The two lba->logical calculations are inconsistent!
      $lsectstr = "lsect [SR] or $lsect2 [fdisk]";
      if (!$already_warned1) {
         print "Warning: SpinRite says LBA $sect is logical sector $lsect, but\n" .
               "fdisk says it is logical sector $lsect2.  Results might be wrong.\n";
         $already_warned1 = 1;
      }
   }
   # based on the cluster size, we skip additional nfi commands for the same cluster:
   if ((-1 != $lsect) && (-1 != $prev_lsect) && (($lsect-$prev_lsect) < $clustersize)) {
      $c1 = int( $prev_lsect / $clustersize );
      $c2 = int( $lsect / $clustersize );
      if ($c1 == $c2) {
         next;  # no need to do another nfi command for another sector in the same cluster
      }
   }
   # write the nfi command:
   if (!$noREMs) {
      printf NFICMDS "Rem LBA sect $sect, byte ofs 0x%03X00, logical sect $lsectstr\r\n", $sect*2;
   }
   if ((-1 eq $lsect) || ('?' eq $drvltr) || $physicalmode) {
      print NFICMDS "nfi \\device\\harddisk0\\dr0 $sect\r\n";
      # Note: use dmdiag.exe to find actual drive spec, if drive under test is not 1st/primary drive in the computer
   } else {
      print NFICMDS "nfi $drvltr: $lsect\r\n";
   }
   if (!$noREMs) {
      print NFICMDS "\r\n";
   }
   $count_actual_num_nfi_cmds++;
   $prev_lsect = $lsect;
}
close NFICMDS;
print "Created '$nficmds_bat' with $count_actual_num_nfi_cmds nfi commands.\n";

__END__
