#!/usr/bin/perl -w
#
# Note: the '-w' (warnings) option also ensures that this script will run
# under Linux even with cr+lf line termination, because it prevents the CR
# from immediately following the "/usr/bin/perl".  (Tricky, eh?)
# For Perl 5, you could, instead, use "-X" (disable warnings), but that's
# incompatible with Perl 4, and is generally a bad idea, anyhow.

$| = 1;

# ddrlognot.pl   ("DDR LOGfile logical .NOT.")
#
# This is a Perl program which reads a GNU ddrescue 1.x "log file" (really,
# a sector list file), and creates from it an "inverted" or "negated"
# version, in which all the '-' (bad/unrecovered) sectors are listed as
# '+' (good/recovered), and vice-versa.
#
# By default, '/', '?' & '*' sectors are treated as '-' (bad/unrecovered),
# so such sectors will be listed as '+' in the output file.  You may
# add the '-g' (good) option to treat them as '+' (good/recovered),
# or you may add the '-k' (keep-as-is) option to leave them as-is in
# the output file.
#
#
# BTW, here's a hint.  Open a console and do:
#   tail -s 3 -f /var/log/messages
# to watch the messages logged during a ddrescue attempt.


# This program is UNCOPYRIGHTED.  It was written by:
#
#    Dave Burton
#    Burton Systems Software
#    Tel: +1 919-481-0149
#    Email: dave at burtonsys.com
#


#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "4, 12-May-15";
# the TLIB Version Control version number and last-modified date


$verbose = 0;      # will be changed if '-v' is specified
$treatasgood = 0;  # will be changed if '-g' is specified
$keepasis = 0;     # will be changed if '-k' is specified


$perlver = "3 or earlier";
if ($] =~ /\$\$Revision\:\s*([0-9.]+)\s/) {
   $perlver = $1;  # probably 4.something
} elsif ($] =~ /([0-9][0-9.]*)/) {
   $perlver = $1;  # probably 5.something or 6.something
}


@sv_ARGV = @ARGV;  # save raw parameter list, for display to the user

# display command-line parameters (just once)
$args_shown = 0;
sub showargs {
   if (!$args_shown) {
      print join( " ", "perl $0", @sv_ARGV ) . "\n";  # echo command-line arguments
      $args_shown = 1;
   }
}


# Handle input parameters and open all the files:

while (($#ARGV >= 0) && ($ARGV[0] =~ /^-([vkg])/)) {
   if ($1 eq 'v') {
      $verbose++;
   } elsif ($1 eq 'k') {
      $keepasis++;
   } elsif ($1 eq 'g') {
      $treatasgood++;
   }
   shift @ARGV;
}
($log_input, $log_output) = @ARGV;

if ($verbose) {
   &showargs;
   print "You are using Perl version $perlver\n";
}

# These are the conversions we'll do:
$cvrt{'+'} = '-';
$cvrt{'-'} = '+';
$cvrt{'/'} = '+';
$cvrt{'?'} = '+';
$cvrt{'*'} = '+';
if ($treatasgood) {
   $cvrt{'/'} = '-';
   $cvrt{'?'} = '-';
   $cvrt{'*'} = '-';
} elsif ($keepasis) {
   $cvrt{'/'} = '/';
   $cvrt{'?'} = '?';
   $cvrt{'*'} = '*';
}


if ((!defined $log_input) || ($keepasis && $treatasgood)) {
   die "ddrlognot.pl   (\"DDRescue LOGfile logical .NOT.\")   v.$ver_ordinal_and_date\n" .
       "\n" .
       "This is a Perl program which reads a GNU ddrescue 1.x \"log file\" (really,\n" .
       "a sector list file), and creates from it an \"inverted\" or \"negated\"\n" .
       "version, in which all the '-' (bad/unrecovered) sectors are listed as\n" .
       "'+' (good/recovered), and vice-versa.\n" .
       "\n" .
       "By default, '/', '?' \& '*' sectors are treated as '-' (bad/unrecovered),\n" .
       "so such sectors will be listed as '+' in the output file.  You may\n" .
       "add the '-g' (good) option to treat them as '+' (good/recovered),\n" .
       "or you may add the '-k' (keep-as-is) option to leave them as-is in\n" .
       "the output file.\n" .
       "\n" .
       "Usage:\n" .
       "   perl -w $0 {options} inputfile outputfile\n" .
       "\n" .
       "{options} can be:\n" .
    #  "   -v  (verbose)\n" .
       "   -k  (keep-as-is, leave '/' and '?' lines alone)\n" .
       "   -g  (treat-as-good, convert '/', '?' and '*' lines to '-' in outputfile)\n" .
       "\n" .
       "E.g., to preserve '/', '?' and '*' lines in outputfile:\n" .
       "   perl -w $0 -k inputfile outputfile\n" .
       "\n" .
       "Note that the parameter order is like 'cp' (input file first), which is\n" .
       "the OPPOSITE of ddrlogand.pl and ddrlogor.pl!\n";
}


&showargs;

open( LOG_INPUT, "<$log_input") || die "ERROR: Could not read $log_input, $!\n";


######################################################################
# Here's a sample of what an old format ddrescue log file looked like:
#
#    # Rescue Logfile. Created by GNU ddrescue version 1.1
#    #      pos        size  status
#    0x00000000  0x0004BE00  +
#    0x0004BE00  0x00000200  -
#    0x0004C000  0x33078E00  +
#    0x330C4E00  0x00001000  -
#    0x330C5E00  0x8ACDF200  +
#
######################################################################


######################################################################
# Here's a sample of what a newer ddrescue log file looks like:
#
#    # Rescue Logfile. Created by GNU ddrescue version 1.11
#    # current_pos  current_status
#    0x244839C00     ?
#    #      pos        size  status
#    0x00000000  0x171CE5000  +
#    0x171CE5000  0x00000200  -
#    0x171CE5200  0x0001B000  *
#    0x171D00200  0x0049AE00  +
#    0x17219B000  0x00000200  -
#    0x17219B200  0x00015000  *
#    0x1721B0200  0xD268A200  +
#    0x24483A400  0x19AE13BC00  ?
#
######################################################################



# Read log_input, and write log_output
$linecounter = 0;
while (<LOG_INPUT>) {
   $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
   if (0 == $linecounter) {
      if ($_ !~ /^\# Rescue/) {
         die "ERROR: '$log_input' is not a ddrescue logfile (doesn't begin with '# Rescue').\n";
      } else {
         open( LOG_OUTPUT, ">$log_output" ) || die "ERROR: Could not create $log_output, $!\n";
         binmode LOG_OUTPUT;  # write output in binary mode, to force LF-only line termination even under DOS/Windows
         if ($_ =~ /^(.*) \(inverted\)$/) {
            $_ = $1;
         } else {
            $_ .= ' (inverted)';
         }
      }
   } elsif ((2 == $linecounter) && ($_ =~ /^0x[0-9A-F]+\s+[\-\+\/\?\*]$/)) {
      # reset current_pos to 0
      $_ = "0x000000000     ?";
   } else {
      if ($_ =~ /^(\d.+)([\-\+\/\?\*])$/) {
         $_ = $1 . $cvrt{$2};
      }
   }
   print LOG_OUTPUT "$_\n";
   $linecounter++;
}

__END__

