#!/usr/bin/perl

$| = 1;

# ddrsplit.pl   ("DDRescue logfile SPLITter")
#
# This is a Perl program which reads a GNU ddrescue 1.x "log file" (really,
# a sector list file), and splits it into multiple smaller log files,
# representing smaller sections of a failing hard disk drive.
#
# Each generated small log file represents a section (sector range) of the
# drive.  In each small log file, the other sectors (that is, the sectors
# outside the sector range represented by that small log file), are marked "+"
# to prevent ddrescue from reading them when using that small log file.
#
# You can use ddrescue with any of the smaller log files, to process the
# smaller sections of the failing hard disk drive.  You can then recombine
# the smaller log files using the ddrcombine.pl program.
#
# The main reason for doing all this is speed.  ddrescue bogs down badly when
# its logfile gets too large.
#

# This program is UNCOPYRIGHTED.  It was written by:
#
#    Dave Burton
#    Burton Systems Software
#    Tel: +1 919-481-0149
#    Email: dave at burtonsys.com
#
#
# v.1 - 26-Oct-2005 - first version.  -DAB
# v.2 -  6-Nov-2005 - '-m' option, and smarter division into nn parts.  -DAB
# v.3 - 17-Sep-2009 - accomodate the status indicator in the ddrescue logfile.  -DAB

#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "10, 05-Oct-09";
# the TLIB Version Control version number and last-modified date


$verbose = 0;
$testmode = 0;
$shortnames = 1;
$move_inputfile = 0;
$skip_prologue_epilogue = 0;


$perlver = "3 or earlier";
if ($] =~ /\$\$Revision\:\s*([0-9.]+)\s/) {
   $perlver = $1;  # probably 4.something
} elsif ($] =~ /([0-9][0-9.]*)/) {
   $perlver = $1;  # probably 5.something or 6.something
}


@sv_ARGV = @ARGV;  # save raw parameter list, for display to user in '-v' (verbose) mode

# Handle input parameters and open files:

# check for '-v' and/or '-t' and/or '-m' and/or '-o'
while (($#ARGV >= 0) && ($ARGV[0] =~ /^-[vtoms]/)) {
   $arg = shift @ARGV;
   if ($arg =~ /v/) {
      $verbose++;
   }
   if ($arg =~ /t/) {
      $testmode = 1;
   }
   if ($arg =~ /o/) {
      $shortnames = 0;  # "name by Offsets" instead of using short names
   }
   if ($arg =~ /m/) {
      $move_inputfile = 1;
   }
   if ($arg =~ /s/) {
      $skip_prologue_epilogue = 1;
   }
}

if ($#ARGV < 1) {
   die "ddrsplit.pl   (\"DDRescue logfile SPLITter\")   v.$ver_ordinal_and_date\n" .
       "\n" .
       "Usage (to divide inputfile into nn sections):\n" .
       "   perl -w ddrsplit.pl inputfile + nn\n" .
       "   (nn should be 2..200)\n" .
       "\n" .
       "or, to divide it at specified byte offsets (which may use 0x... syntax):\n" .
       "   perl -w ddrsplit.pl inputfile byteoffset1 byteoffset2 ...\n" .
       "\n" .
       "or, to divide it at specified LBA sector numbers use the ddrescue-style 'b' suffix:\n" .
       "   perl -w ddrsplit.pl inputfile {decimal1stLBAno}b {decimal2ndLBAno}b ...\n" .
       "   (yes, I realize 'b' for 'block' instead of 's' for 'sector' is counter-intuitive)\n" .
       "\n" .
       "or, for equal-sized sections of nnn bytes each:\n" .
       "   perl -w ddrsplit.pl inputfile nnn +\n" .
       "\n" .
       "or, for equal-sized sections of {nnn} sectors each (where {nnn} must be decimal):\n" .
       "   perl -w ddrsplit.pl inputfile {nnn}b +\n" .
       "\n" .
       "For verbose output, specify -v:\n" .
       "   perl -w ddrsplit.pl -v inputfile etc...\n" .
       "\n" .
       "To just test, but not actually write any output files, specify -t:\n" .
       "   perl -w ddrsplit.pl -t inputfile etc...\n" .
       "\n" .
       "To move the input logfile to directory 'unused/', specify -m:\n" .
       "   perl -w ddrsplit.pl -m inputfile etc...\n" .
       "\n" .
       "To skip the '+' prologue and epilogue in each output file, specify -s:\n" .
       "   perl -w ddrsplit.pl -s inputfile etc...\n" .
       "\n" .
       "Normally, output file names are generated from the input file name plus section\n" .
       "number.  For example, if the command is:\n" .
       "   perl -w ddrsplit.pl drive.log 0x600000000 0xC00000000\n" .
       "or equivalently:\n" .
       "   perl -w ddrsplit.pl drive.log 50331648b 100663296b\n" .
       "then there will be three output files, and their names will be:\n" .
       "   drive.log-01, drive.log-02, drive.log-03\n" .
       "\n" .
       "To name the output files by starting byte Offsets, specify -o, like this:\n" .
       "   perl -w ddrsplit.pl -o drive.log 0x600000000 0xC00000000\n" .
       "then the three output files will be named:\n" .
       "   drive.log-000000000000, drive.log-000600000000, drive.log-000C00000000\n" .
       "\n" .
     #  "'-v' can be repeated for very verbose output.\n" .
     #  "\n" .
       "Byte offsets and byte sizes should be multiples of 512 bytes (0x200).\n" .
       "\n" .
       "More detailed instructions are included in the source code comments,\n" .
       "at the beginning of ddrsplit.pl.\n";
}

print join( " ", $0, @sv_ARGV ) . "\n";  # echo command-line arguments
if ($verbose) {
   print "You are using Perl version $perlver\n";
}

$log_input = shift @ARGV;
open( LOG_INPUT, "<$log_input") || die "ERROR: Could not read $log_input, $!\n";


######################################################################
# Here's a sample of what a ddrescue log file looks like; line 3 is the
# status line; lines 2 & 3 were introduced in ddrescue 1.6:
#
#    # Rescue Logfile. Created by GNU ddrescue version 1.9
#    # current_pos  current_status
#    #    0x00000000     ?
#    #      pos        size  status
#    0x00000000  0x0004BE00  +
#    0x0004BE00  0x00000200  -
#    0x0004C000  0x33078E00  +
#    0x330C4E00  0x00001000  -
#    0x330C5E00  0x8ACDF200  +
#
######################################################################


# input is an integer represented as a string (probably hexadecimal preceeded by '0x'),
# result is the integer divided by 512.
sub divby512 {
   local($dividend) = @_;
   # print "dividend=$dividend\n";
   local($result) = 0;
   if ($dividend =~ /^0x/) {
      # print "tricky way!\n";
      # tricky divide by 512: drop the last 2 hexidecimal digits, and divide by 2 (which is more robust for very large integers)
      $result = substr($dividend,0,length($dividend)-2);
      $result = int(eval($result) / 2);  # gotta use eval() because it starts with 0x
   } else {
      # print "conventional way,\n";
      $result = int((0+$dividend) / 512);
   }
   # print "result=$result\n";
   return $result;
}



# Internal representation of the sector list is a bit tricky.
# The problem is that there are soooo many sectors.  A 300 GB
# disk drive has 600 million sectors.  Even at one character
# per sector, that's too big to represent in memory as a simple
# array or string.  So we have to represent it like it is in
# the sector list file, as a list of ranges.  But that means
# that we can't use a simple array or hash lookup to check the
# status of a particular sector; we'll have to call a function
# to do a search.


if ($verbose) {
   print "reading $log_input . . .\n";
}


# First, read log_input (the ddrescue sector list input file)
$numranges = 0;
$lastline = '';
$highest_from_inp1 = 0;
while (<LOG_INPUT>) {
   if (($_ =~ /\d/) && ($_ !~ /^\#/)) {
      $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
      # Parse a ddrescue log file line:
      @tmp = split( /\s+/, $_ );
      # for ($i=0; $i<=$#tmp; $i++) { print "dbg: tmp[$i]='" . $tmp[$i] . "'\n"; }
      if ((1 == $#tmp) && ($tmp[1] =~ /^[\?\*\/\-\+]$/)) {
         # the new status line (introduced in ddrescue 1.6)
         next;
      }
      ($pos_byt, $len_byt, $status) = @tmp;
      # if ($verbose) {
      #    print "dbg: pos_byt='$pos_byt'  len_byt='$len_byt'  status='$status'\n";
      # }
      if (($#tmp != 2) || ($pos_byt !~ /\d/) || ($len_byt !~ /\d/)) {
         print "ERROR: unrecognized line in '$log_input':\n$_\n";
         next;
      }
      $lastline = '';
      if ($len_byt eq '-1') {
         print "Warning: length=-1 for status = '$status' in '$log_input', line ignored:\n$_\n";
         next;
      }

      # Now convert byte offsets and counts to LBA sector number & sector count:
      $pos_sect = &divby512($pos_byt);
      $len_sect = &divby512($len_byt);

      # print "dbg: Bad sect no. = $pos_sect, count = $len_sect, '$status'\n";
      # for ($i=0; $i<$len_sect; $i++) {
      #   $originally_bad_sectors{$pos_sect} = 1;
      #   $num_originally_bad_sectors++;
      #   $pos_sect++;
      # }

      $offsets[$numranges] = $pos_sect;
      $lengths[$numranges] = $len_sect;
      $stati[$numranges] = $status;  # '-' or '+' or '/' or '?'
      $numranges++;
      $highest_from_inp1 = $pos_sect + $len_sect;
      if ($status eq '') { print "gak1\n";}  # dbg
   }
}


$count_rescued1 = $count_failed1 = 0;
$first_unrescued = $highest_from_inp1;;
$end_of_unrescued_range = 0;
for ($i=0; $i<$numranges; $i++) {
   $status = $stati[$i];
   # $ofs = $offsets[$i];
   $len = $lengths[$i];
   if ('+' eq $status) {
      $count_rescued1 += $len;
   } else {
      $ofs = $offsets[$i];
      $endrange = $ofs + $len;
      if ('-' eq $status) {
         $count_failed1 += $len;
      }
      if ($ofs < $first_unrescued) {
         $first_unrescued = $ofs;
      }
      if ($endrange > $end_of_unrescued_range) {
         $end_of_unrescued_range = $endrange;
      }
   }
}
$count_untried1 = $highest_from_inp1 - ($count_rescued1 + $count_failed1);
print "'$log_input' describes $highest_from_inp1 sectors  ($count_rescued1 rescued, $count_failed1 failed, $count_untried1 untried).\n";
if ($first_unrescued > $end_of_unrescued_range) {
   die "ERROR: all sectors named in '$log_input' were already rescued; aborting!\n";
}
if ($first_unrescued == $end_of_unrescued_range) {
   die "ERROR: all but 1 sector named in '$log_input' were already rescued; aborting!\n";
}
$max_no_pieces = $end_of_unrescued_range - $first_unrescued;
if ($max_no_pieces > 200) {
   $max_no_pieces = 200;  # we won't split it into more than 200 little log files
}

# Now figure out the split points.  There are three ways to specify them:
#
#    For specific split points:
#      perl -w ddrsplit.pl inputfile byteoffset1 byteoffset2 ...
#
#    For equal-sized sections of nn bytes each (nn can be decimal or 0x... for hex)
#      perl -w ddrsplit.pl inputfile nn +
#    (this divides the entire range of the input file)
#
#    For nn aproximately equal-sized sections (nn must be a decimal integer greater than 1):
#      perl -w ddrsplit.pl inputfile + nn
#    (this divides the unrescued range, so it can be used to re-divide a log file
#    which represents just one piece of the disk being covered; that is, it can
#    be used to further subdivide one of the log files previously output by this
#    same program)
#
# Note that inputfile and switches (if any) have already been removed from @ARGV.
#

if (0 == $#ARGV) {
   if ($ARGV[0] =~ /^\+[0-9]/) {
      # special case handling for when the user left out the space after the '+'
      # change "perl ddrsplit.pl drive.log +5" to "perl ddrsplit.pl drive.log + 5"
      $ARGV[1] = substr($ARGV[0],1);
      $ARGV[0] = '+';
   } elsif ($ARGV[0] =~ /[0-9A-Fa-f]\+$/) {
      # special case handling for when the user left out the space before the '+'
      # change "perl ddrsplit.pl drive.log 0x50000+" to "perl ddrsplit.pl drive.log 0x50000 +"
      $ARGV[0] = substr( $ARGV[0], 0, length($ARGV[0])-1 );
      $ARGV[1] = '+';
   }
}
if ((1 == $#ARGV) && ('+' eq $ARGV[0])) {
   # nn equal-sized sections
   $num_pieces = 0+$ARGV[1];
   if (($num_pieces < 2) || ($max_no_pieces > 200)) {
      if ($max_no_pieces < 200) {
         die "ERROR: The no. of parts into which an input log file may be divided is 2..200,\n" .
             "generally, but 2..$max_no_pieces for '$log_input'.\n";
      } else {
         die "ERROR: The no. of parts into which the input log file is divided must be 2..200.\n";
      }
   }

   # The range we divy up is not the entire disk, it is:
   #  $first_unrescued .. $end_of_unrescued_range
   # so that you can easily use this program to split up
   # one of the pieces which resulted from a previous run
   # of this program.

   $siz_sect = int( (($end_of_unrescued_range-$first_unrescued)+($num_pieces-1)) / $num_pieces );
   @split_points = (0);
   $splitpoint = $first_unrescued + $siz_sect;
   for ($i=1; $i<$num_pieces; $i++) {
      push( @split_points, $splitpoint );
      $splitpoint += $siz_sect;
   }
} elsif ((1 == $#ARGV) && ('+' eq $ARGV[1])) {
   # Divide up into equal-sized sections of specified size.
   # The range we divy up is the entire drive.
   $arg = $ARGV[0];
   if (($arg =~ /b$/) && ($arg !~ /^0x/)) {
      # specified in sectors, a decimal integer with the 'b' suffix
      $siz_sect = 0 + substr( $arg, 0, length($arg)-1 );
      $siz_bytes = 512 * siz_sect;
   } else {
      $siz_bytes = $arg;
      $siz_sect = &divby512( $siz_bytes );
   }
   if ($siz_sect < 1) {
      die "ERROR: piece size (\"$arg\") must be at least 1 sector\n" .
          "(one sector is 512=0x200 bytes).\n";
   }
   $num_pieces = int( ($highest_from_inp1+($siz_sect-1)) / $siz_sect );
   if (($num_pieces < 2) || ($num_pieces > 200)) {
      die "ERROR: a piece size of $siz_bytes would divide '$log_input' into $num_pieces pieces.\n" .
          "The number of pieces into which the input log file is divided must be 2..200.\n";
   }
   $splitpoint = 0;
   for ($i=0; $i<$num_pieces; $i++) {
      push( @split_points, $splitpoint );
      $splitpoint += $siz_sect;
   }
} else {
   # specified specific split points, specified in bytes or sectors:
   $num_pieces = 2 + $#ARGV;
   $siz_sect = 0; # varies
   @split_points = (0);
   foreach $arg (@ARGV) {
      if (($arg =~ /b$/) && ($arg !~ /^0x/)) {
         # specified in sectors, a decimal integer with the 'b' suffix
         $arg = 0 + substr( $arg, 0, length($arg)-1 );
         push( @split_points, $arg );
      } else {
         # specified in bytes
         push( @split_points, &divby512($arg) );
      }
   }
   $prev = 0;
   foreach $splitpoint (@split_points) {
      if ($splitpoint < $prev) {
         $tmp = sprintf( "0x%06X00, 0x%06X00", (2*$prev), (2*$splitpoint) );
         die "ERROR: split points must be monotonically increasing, by at least 512 bytes;\n $tmp\n";
      }
      if ($splitpoint >= $highest_from_inp1) {
         $tmp = sprintf( "0x%06X00 >= 0x%06X00", (2*$splitpoint), (2*$highest_from_inp1) );
         die "ERROR: split points may not exceed the input disk size;\n $tmp\n";
      }
      $prev = $splitpoint;
   }
}
if ($verbose) {
   print "The specified parameters will divide '$log_input' into $num_pieces smaller files";
   if (0 == $siz_sect) {
      print ".\n";
   } else {
      printf ",\neach representing at most %d sectors (0x%02X00 bytes).\n", $siz_sect, 2 * $siz_sect;
   }
}

if (($#split_points+1) != $num_pieces) {
   die "ERR: num_pieces/1  #split_points=$#split_points, num_pieces=$num_pieces\n";
}
if ($verbose >= 2) {
   printf "dbg: #split_points=%d, split_points[#split_points]=%s, highest_from_inp1=%d\n", $#split_points, $split_points[$#split_points], $highest_from_inp1;
}
if ($split_points[$#split_points] > $highest_from_inp1) {
   die "ERR: num_pieces/2\n";
}


# Calculate the name of an output file.  We support two ways, controlled by the -s option
sub outfilename {
   local( $index ) = @_;
   local( $splitpoint, $result );
   if ($shortnames) {
      if ($num_pieces > 99) {
         $fmt = "%s-%03d";
      } else {
         $fmt = "%s-%02d";
      }
      $result = sprintf( $fmt, $log_input, $index+1 );
   } else {
      $splitpoint = $split_points[$index];
      $result = sprintf( "%s-%010X00", $log_input, $splitpoint );
   }
   return $result;
}


if ($testmode || $verbose) {
   print "The $num_pieces output file names are:\n";
   for ($i=0; $i<$num_pieces; $i++) {
      print &outfilename($i) . "\n";
   }
}

if ($verbose) {
   print "dbg: The $num_pieces sections are:\n";
}
for ($i=0; $i<$num_pieces; $i++) {
   $begin_section = $split_points[$i];
   if ($i == ($num_pieces-1)) {
      $end_section = $highest_from_inp1;
   } else {
      $end_section = $split_points[$i+1];
   }
   $len_section = $end_section - $split_points[$i];
   $section_len[$i] = $len_section;
   if ($verbose) {
      printf "dbg: [$i]  0x%06X00-0x%06X00  (len=%d=0x%07X sectors, = 0x%06X00 bytes)\n", (2*$begin_section), (2*$end_section), $len_section, $len_section, (2*$len_section);
   }
}

if ($testmode) {
   exit( 0 );
}


# Format a ddrescue logfile line.  Input parameters are in sectors (512 bytes),
# but the result is displayed in bytes.
sub displayrange {
   local( $ofs, $len, $status ) = @_;
   if (!defined $status) {
      $status = 'undef';  # should never happen
   }
   local( $result ) = sprintf( "0x%06X00  0x%06X00  %s", (2*$ofs), (2*$len), $status );
   return $result;
}


# for each output file to be generated...
for ($section=0; $section<$num_pieces; $section++) {
   if ($verbose) {
      print "dbg: section no. " . ($section+1) . " (of $num_pieces)\n";
   }
   $splitpoint = $split_points[$section];
   $len_section = $section_len[$section];
   $end_section = $splitpoint + $len_section;
   $log_output = &outfilename( $section );

   if (!open(LOG_OUTPUT,">$log_output")) {
      die "ERROR: Could not create output file '$log_output', $!\n";
   }
   binmode LOG_OUTPUT;  # write output in binary mode, to force LF-only line termination even under DOS/Windows
   print "writing '$log_output' ";
   print LOG_OUTPUT "# Rescue Logfile. Created by GNU ddrescue version 1.9-burtonhack1\n" .
         "# current_pos  current_status\n" .
         "0x00000000     ?\n" .
         "#      pos        size  status\n";
   if (($splitpoint > 0) && !$skip_prologue_epilogue) {
      # print "dbg: skip_prologue_eqilogue=$skip_prologue_epilogue\n";
      print LOG_OUTPUT &displayrange(0,$splitpoint,'+') . "\n";
   }
   $count_failed2 = $count_untried2 = 0;
   for ($i = 0; $i < $numranges; $i++) {
      $ofs = $offsets[$i];
      $len = $lengths[$i];
      $status = $stati[$i];
      if (($ofs+$len) <= $splitpoint) {
         next;  # entire interval preceeds the sectors in this output file's section
      }
      if ($ofs >= $end_section) {
         last;  # entire interval is past the end of the section
      }
      # There are four more cases:
      # 1) The interval is entirely contained in the range of sectors in this section
      # 2) The interval begins before the section, and ends within it
      # 3) The interval begins within the section, and ends past it
      # 4) The interval contains the entire section
      if ($ofs < $splitpoint) {
         # interval begins before the section's sector range; shorten the interval
         $len -= ($splitpoint - $ofs);
         $ofs = $splitpoint;
      }
      if (($ofs+$len) > $end_section) {
         # interval ends after the section's sector range; shorten the interval
         $len = ($end_section-$ofs);
      }
      print LOG_OUTPUT &displayrange($ofs,$len,$status) . "\n";
      if ('-' eq $status) {
         $count_failed2 += $len;
      } elsif ('+' ne $status) {
         # '/' or '?'
         $count_untried2 += $len;
      }
   }
   if (($highest_from_inp1 > $end_section) && !$skip_prologue_epilogue) {
      print LOG_OUTPUT &displayrange( $end_section, $highest_from_inp1-$end_section, '+' ) . "\n";
   }
   close LOG_OUTPUT;
   print " ($count_failed2 failed, $count_untried2 untried).\n";
}#for each output file to be generated

# '-m' option
if ($move_inputfile) {
   @tmp = split( '/', $log_input );
   splice( @tmp, $#log_input, 0, 'unused' );
   $dest = join( '/', @tmp );
   pop( @tmp );
   $unuseddir = join( '/', @tmp );
   print "mkdir \"$unuseddir\"\n";
   if (!mkdir( $unuseddir, 0x777 )) {
      # printf "dbg: errno=%d='%s'\n", $!, $!;  # 17='File exists'
      if ($verbose || ($! !~ /exists/)) {
         if ($! !~ /exists/) {
            print "ERROR: ";
         }
         printf "'mkdir \"%s\"' failed, errno=%d='$!'\n", $unuseddir, $!, $!;
      }
   }
   # `echo mv $log_input unused/\n`;
   print "mv -f \"$log_input\" \"$dest\"\n";
   if (!rename( $log_input, $dest )) {
      print "ERROR: 'mv -f \"log_input\" \"dest\"' failed, $!\n";
   }
}

print "\n";

__END__

