#!/usr/bin/perl

$| = 1;

# nficruncher.pl   ("NTFS File sector Information tool output cruncher")
#
# This is a Perl program which reads the output of a series of NFI commands,
# and from then deduces which (presumably failing) sectors are important and
# which are unimportant.
#
# The unimportant ones need not be rescued.  The important ones might merit
# the "freezer treatment" or SpinRite.
#
#
# I plan to turn this into a specialized variant of ddrsplit.pl, which splits
# a ddrescue log file into "important" and "unimportant" pieces.
#

# This program is UNCOPYRIGHTED.  It was written by:
#
#    Dave Burton
#    Burton Systems Software
#    Tel: +1 919-481-0149
#    Email: dave at burtonsys.com
#


#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "8, 30-Mar-08";
# the TLIB Version Control version number and last-modified date


$rc = 0;  # no errors so far

$verbose = 0;             # '-v' sets it to 1, -vv or "-v -v" sets it to 2
$report_damaged = 0;      # '-d' sets it to 1, '-r' sets it to 2
$report_important = 0;    # '-i' sets it to 1
$report_unimportant = 0;  # '-u' sets it to 1
$cluster_size = 0;        # '-cNN' sets it to NN, where NN must be in {1,2,4,8,16,32,64,128}
$partition_offset = 0;    # '-pNNNN' sets it to NNNN, the offset (in sectors) of the NTFS partition
$partition_offset_specified = 0;
$disk_size = 0;           # '-sMMMM' sets it to MMMM, the size (in sectors) of the entire disk drive
$disk_size_specified = 0;
$fdisk_out = '';          # '-ffilename' where filename is the fdisk -lu output, an alternative to -p and -s


$perlver = "3 or earlier";
if ($] =~ /\$\$Revision\:\s*([0-9.]+)\s/) {
   $perlver = $1;  # probably 4.something
} elsif ($] =~ /([0-9][0-9.]*)/) {
   $perlver = $1;  # probably 5.something or 6.something
}


@sv_ARGV = @ARGV;  # save raw parameter list, for display to the user

# check for '-v' and/or '-d' and/or '-r' and/or '-i' and/or '-u' and/or '-cNN'
while (($#ARGV >= 0) && ($ARGV[0] =~ /^-[vdiucprsf]/)) {
   if ($ARGV[0] =~ /^-c(\d+)$/) {
      # '-cNN' to set cluster size
      $newclustersize = 0+$1;
      # check for illegal sectors-per-cluster value:
      $badspc = 1;
      foreach $spc (1,2,4,8,16,32,64,128) {
         if ($newclustersize == $spc) {
            $badspc = 0;
         }
      }
      if ($badspc) {
         print STDERR "ERROR: '$ARGV[0]' specifies an impossible cluster size.\n" .
               "The cluster size should be in {1,2,4,8,16,32,64,128}.\n";
         exit(2);
      } else {
         if ($verbose) {
            print "dbg: cluster size set to $newclustersize sectors/cluster.\n";
         }
         $cluster_size = $newclustersize;
      }
   } elsif ($ARGV[0] =~ /^-p(\d+)$/) {
      # '-pXXX' to set partition 'start' sector for the NTFS partition
      $partition_offset = 0+$1;
      $partition_offset_specified = 1;
      if ($partition_offset < 63) {
         print "Warning: $partition_offset is an unlikely NTFS partition 'start' sector!\n";
      }
   } elsif ($ARGV[0] =~ /^-s(\d+)$/) {
      # '-sYYY' to set partition size for the NTFS partition
      $disk_size = 0+$1;
      $disk_size_specified = 1;
      if ($disk_size < 1200) {
         print "Warning: a disk size of $disk_size sectors is unusually small!\n";
      }
   } elsif ($ARGV[0] =~ /^-f([\w\-\+\.\$\(\)]*)$/) {
      # '-ffilename' to specify a file containing 'fdisk -lu' output
      $fdisk_out = $1;
      if ('' eq $fdisk_out) {
         $fdisk_out = 'fdisk-lu_out.txt';  #default
      }
   } else {
      foreach $ch (split(//,$ARGV[0])) {
         if ('v' eq $ch) {
            $verbose++;
         } elsif ('d' eq $ch) {
            $report_damaged |= 1;
         } elsif ('r' eq $ch) {
            $report_damaged |= 2;
         } elsif ('i' eq $ch) {
            $report_important++;
         } elsif ('u' eq $ch) {
            $report_unimportant++;
         } elsif ('f' eq $ch) {
            if ('' eq $fdisk_out) {
               $fdisk_out = 'fdisk-lu_out.txt';  #default
            }
         } elsif ('-' ne $ch) {
            print "Warning: unrecognized option (\"-$ch\") ignored.\n";
         }
         # Note: if you add another option, be sure to fix the 'while' condition, above
      }
   }
   shift @ARGV;
}


# give usage help and end program
sub givehelp {
   print "nficruncher.pl     v.$ver_ordinal_and_date\n" .
         "(\"NTFS File sector Information tool output cruncher\")\n" .
         "\n" .
         "The assumption is that you used ddr2nfi.pl to generate a Windows NT/2K/XP/2003\n" .
         "batch file (perhaps called nficmds.bat), containing NFI commands for the\n" .
         "unrecovered sectors.  This Perl program processes the concatinated output of\n" .
         "those NFI commands.\n" .
         "\n" .
         "Usage:\n" .
         "   nficmds.bat >nfioutput.txt\n" .
         "   perl -w $0 {options} nfioutput.txt\n" .
         "\n" .
         "where {options} can be:\n" .
         "   -v     (verbose debug prints)\n" .
         "   -d     (damaged file report)\n" .
         "   -r     (generate .bat file to rename damaged files)\n" .
         "   -i     (important sectors report, in ddrescue logfile format)\n" .
         "   -u     (unimportant sectors report, in ddrescue logfile format)\n" .
         "   -cNN   (assume cluster size is NN sectors/cluster; default is 8)\n" .
         "   -pXXX  (NTFS partition starts at sector XXX, as reported by 'fdisk -lu')\n" .
         "   -sYYY  (NTFS partition size, in sectors)\n" .
         "   -f     (read 'fdisk-lu_out.txt', the output of Linux 'fdisk -lu ...')\n" .
         "   -ffilename  (read filename', the output of Linux 'fdisk -lu ...')\n" .
         "\n" .
         "Note #1:  You can use clustersize.pl to discover the cluster size and starting\n" .
         "sector number of the NTFS partition.\n" .
         "\n" .
         "Note #2:  If you specify -i or -u, $0 needs to know the starting sector of\n" .
         "the NTFS partition (and, for -i, the disk size).  If there is only one NTFS\n" .
         "partition, you can use the '-f' option, otherwise use '-p' and '-s'.\n";
}


# display command-line parameters (just once)
$args_shown = 0;
sub showargs {
   if (!$args_shown) {
      print join( " ", "perl $0", @sv_ARGV ) . "\n";  # echo command-line arguments
      $args_shown = 1;
   }
}

if ($#ARGV < 0) {
   $rc = 1;
}

if ($verbose) {
   &showargs;
   print "You are using Perl version $perlver\n";
}
if (   (!($report_damaged || $report_important || $report_unimportant ))
    && (-1 != $#sv_ARGV) ) {
   print "ERROR: you should specify an option:  -d, -r, -i, or -u\n";
   sleep(1);
   $rc = 1;
}

if (($partition_offset_specified && $disk_size_specified) && ('' ne $fdisk_out)) {
   print "ERROR: you should either specify '-p' and '-s', or else specify '-f', not both.\n";
   sleep(1);
   $rc = 1;
}

if ($rc) {
   &givehelp;
   exit(1);
} else {
   &showargs;
}

if ('' ne $fdisk_out) {
   open( FDISK_OUT, "<$fdisk_out") || die "ERROR: Could not read $fdisk_out, $!\n";
   # Read FDISK_OUT, to get the location and size of the NTFS partition.  Here's an example
   # (note the blank line at the beginning):
   #----------------------------------------------------------------------
   #
   # Disk /dev/hdg: 160.0 GB, 160000000000 bytes
   # 255 heads, 63 sectors/track, 19452 cylinders, total 312500000 sectors
   # Units = sectors of 1 * 512 = 512 bytes
   #
   #    Device Boot      Start         End      Blocks   Id  System
   # /dev/hdg1              63      112454       56196   de  Dell Utility
   # /dev/hdg2   *      112455   305154674   152521110    7  HPFS/NTFS
   # /dev/hdg3       305154675   312496379     3670852+  db  CP/M / CTOS / ...
   #----------------------------------------------------------------------
   $saw_Units_line = $num_partitions = 0;
   while (<FDISK_OUT>) {
      $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
      if ($_ eq 'Units = sectors of 1 * 512 = 512 bytes') {
         $saw_Units_line = 1;
      }
      # match a line like this:
      # 255 heads, 63 sectors/track, 19452 cylinders, total 312500000 sectors
      elsif ($_ =~ /.*total\s+(\d+)\s+sectors/) {
         $total_sectors = 0+$1;
         if ($disk_size_specified) {
            if ($disk_size == $total_sectors) {
               print "Note: you needn't specify both '-f' and '-s' parameters.\n";
            } else {
               print "Warning: disk size from '$fdisk_out' is $total_sectors, but you used the\n" .
                     "'-s' parameter to force $0 to pretend the disk size is $disk_size.\n";
            }
         } else {
            $disk_size = 0+$1;
         }
      }
      # match a line like this:
      # '/dev/hdg2   *      112455   305154674   152521110    7  HPFS/NTFS'
      elsif ($_ =~ /^(\/dev\/\w*[a-z])(\d+)\s+(\*|)\s*(\d+)\s+(\d+)\s+(\d+)(\+|)\s+(\w{1,2})\s*(.*)$/) {
         ($dev, $partnum, $bootind, $start, $end, $blocks, $halfblockind, $id, $system) =
           ($1, $2, $3, $4, $5, $6, $7, $8, $9);
         if ($verbose) {
            print "dbg: device= $dev$partnum\n";
            print "     bootind=$bootind\n";
            print "     start=$start\n";
            print "     end=$end\n";
            print "     blocks=$blocks";
            if ('' ne $halfblockind) {
               print ".5";
            }
            print "\n     id=$id\n";
            print "     system=$system\n";
         }
         if ((!$partition_offset_specified) && ($system =~ /NTFS/)) {
            if ($partition_offset > 0) {
               print "Warning: this drive contains more than 1 NTFS partition.  We assume the first one is the correct one.\n";
            } else {
               $partition_offset = $start;
            }
         }
         $num_partitions++;
      }
   }
   close FDISK_OUT;
}

if ($report_important && (($partition_offset <= 0) || ($disk_size <= 0))) {
   print "ERROR: the '-i' option requires that you also specify either '-f' or else both '-p' and '-s'.\n";
   sleep(1);
   &givehelp;
   exit(1);
}

if ($report_unimportant && ($partition_offset < 0)) {
   print "ERROR: the '-u' option requires that you also specify either '-f' or '-p'.\n";
   sleep(1);
   &givehelp;
   exit(1);
}

$from_nfi = shift @ARGV;
open( FROM_NFI, "<$from_nfi") || die "ERROR: Could not read $from_nfi, $!\n";


if (($report_important || $report_unimportant) && !$partition_offset) {
    print "\n" .
          "ERROR: you must also specify the '-pXXX' (partition start) parameter along\n" .
          "with the '-u' or '-i' parameter.\n" .
          "\n";
    &givehelp;
    exit(1);
}


print "Cluster size is ";
if ($cluster_size <= 0) {
   print "assumed to be the default, which is ";
   $cluster_size = 8;  # default cluster size is 8 sectors, since that is how most NTFS partitions are formatted
}
printf "$cluster_size sectors (%d bytes).\n", 512 * $cluster_size;
if ($partition_offset > 0) {
   print "partition_offset=";
   if ($partition_offset > 0) {
      print $partition_offset;
   } else {
      print "unspecified";
   }
   print ", disk_size=";
   if ($disk_size > 0) {
      print $disk_size;
   } else {
      print "unspecified";
   }
   print "\n";
}


# Here are samples of four kinds of NFI output:
#
#
# First, a "normal" file (in this case not very important):
# - - - - - - - -( begin NFI output)- - - - - - - -
# C:\somedir>nfi H: 409016
# NTFS File Sector Information Utility.

# Copyright (C) Microsoft Corporation 1999. All rights reserved.
#
#
# ***Logical sector 409016 (0x63db8) on drive H is in file number 92222.
# \System Volume Information\_restore{B37680B2-BA0A-4E5D-BF30-83E44C588624}\RP873\change.log.1
#     $STANDARD_INFORMATION (resident)
#     $FILE_NAME (resident)
#     $FILE_NAME (resident)
#     $DATA (nonresident)
#         logical sectors 409008-409047 (0x63db0-0x63dd7)
#         logical sectors 950936-950975 (0xe8298-0xe82bf)
#         logical sectors 951920-951959 (0xe8670-0xe8697)
#         logical sectors 16045216-16045223 (0xf4d4a0-0xf4d4a7)
#
# - - - - - - - - ( end NFI output) - - - - - - - -
#
#
#
# Next, an unused cluster (completely unimportant):
# - - - - - - - -( begin NFI output)- - - - - - - -
# C:\somedir>nfi H: 967280
# NTFS File Sector Information Utility.
# Copyright (C) Microsoft Corporation 1999. All rights reserved.
#
#
# ***Logical sector 967280 (0xec270) on drive H is not in any file.
#
# - - - - - - - - ( end NFI output) - - - - - - - -
#
#
#
# Next, a sector in the Master File Table (very important):
# - - - - - - - -( begin NFI output)- - - - - - - -
# C:\rescue\patronstron>nfi H: 6546640
# NTFS File Sector Information Utility.
# Copyright (C) Microsoft Corporation 1999. All rights reserved.
#
#
# ***Logical sector 6546640 (0x63e4d0) on drive H is in file number 0.
# Master File Table ($Mft)
#     $STANDARD_INFORMATION (resident)
#     $FILE_NAME (resident)
#     $DATA (nonresident)
#         logical sectors 6539480-6564583 (0x63c8d8-0x642ae7)
#         logical sectors 16196368-16355007 (0xf72310-0xf98ebf)
#         logical sectors 47387448-47388159 (0x2d31338-0x2d315ff)
#     $BITMAP (nonresident)
#         logical sectors 6539472-6539479 (0x63c8d0-0x63c8d7)
#         logical sectors 6216824-6216831 (0x5edc78-0x5edc7f)
#         logical sectors 14152912-14152919 (0xd7f4d0-0xd7f4d7)
#
# - - - - - - - - ( end NFI output) - - - - - - - -
#
#
#
# Last, a "lost" file.  This might be an artifact of $Mft damage:
# - - - - - - - -( begin NFI output)- - - - - - - -
# C:\rescue\patronstron>nfi H: 28889280
# NTFS File Sector Information Utility.
# Copyright (C) Microsoft Corporation 1999. All rights reserved.
#
#
# ***Logical sector 28889280 (0x1b8d0c0) on drive H is in file number 90164.
# (unknown/unnamed)
#     $STANDARD_INFORMATION (resident)
#     $FILE_NAME (resident)
#     $DATA (nonresident)
#         logical sectors 28889280-28889303 (0x1b8d0c0-0x1b8d0d7)
#
# - - - - - - - - ( end NFI output) - - - - - - - -



$_ = '';
while (<FROM_NFI>) {
   if ($_ =~ /NTFS File Sector Information Utility./) {
      last;
   }
}
if ((!defined $_) || ($_ !~ /NTFS File Sector Information Utility./)) {
   print STDERR "ERROR: '$from_nfi' is not output from NFI.\n";
   exit(1);
}


sub numerically { $a <=> $b; }


# '-d' or '-r'
if ($report_damaged) {
   while (<FROM_NFI>) {
      $_ =~ s/\s*$//g;  # remove trailing whitespace
      if ($_ =~ /\*\*\*Logical sector\s(\d+)\s\(\w*\)\s.*in file number (\d*)\.$/) {
         $sect = $1;
         $fnumber = $2;
         $_ = <FROM_NFI>;  # read file name
         $_ =~ s/\s*$//g;  # remove trailing whitespace
         $fname = $_;
         if (defined $fnum2name{$fnumber}) {
            if ($fname ne $fnum2name{$fnumber}) {
               $oldfname = $fnum2name{$fnumber};
               print STDERR "Warning: multiple file names associated with file number $fnumber.\n" .
                            "Hard links?   Name#1='$oldfname'   (Name#2='$fname'. ignored).\n";
            }
         } else {
            $fnum2name{$fnumber} = $fname;
         }
      } elsif ($_ =~ /\*\*\*Logical sector\s(\d+)\s\(\w*\)\s.*is not in any file\.$/) {
         $unused{$1} = 1;
      }
   }

   if ($report_damaged & 2) {
      open( LOG_OUTPUT, ">rename_damaged.bat" ) || die "ERROR: could not write 'rename_damaged.bat', $!\n";
      binmode LOG_OUTPUT;
      print LOG_OUTPUT "\@echo on\r\n";
      foreach $fname (reverse sort values %fnum2name) {
         # reverse sort so that rename of \abc\xyz happens before rename of \abc
         if ($fname =~ /.*\\([^\\]+)$/) {
            $nameonly = $1;
            print LOG_OUTPUT "ren \"$fname\" \"$nameonly.damaged\"\r\n";
         } else {
            print "Warning: can't rename '$fname'\n";
         }
      }
      close LOG_OUTPUT;
      print "Created 'rename_damaged.bat'\n";
   }
   if ($report_damaged & 1) {
      print "Unused sectors:\n";
      foreach $sect (sort numerically keys %unused) {
         print "$sect\n";
      }
      print "\n" .
            "Damaged files:\n";
      foreach $fname (sort values %fnum2name) {
         print "$fname\n";
      }
   }
}


# Format a ddrescue logfile line.  Input parameters are in sectors (512 bytes),
# but the result is displayed in bytes.
sub displayrange {
   local( $ofs, $len, $status ) = @_;
   if (!defined $status) {
      $status = 'undef';  # should never happen
   }
   local( $result ) = sprintf( "0x%06X00  0x%06X00  %s", (2*$ofs), (2*$len), $status );
   return $result;
}


# Vaguely like the 'iota' function in APL, generate indices from M to N.
# Example:  &range(0,5) returns (0,1,2,3,4,5)
sub range {
   local( $M, $N ) = @_;
   local( @result, $i );
   @result = ();
   for ($i = $M; $i <= $N; $i++) {
      push( @result, $i );
   }
   return @result;
}


# For sorting a set of indices to an array of values, hard-coded to be @offsets;
# this is the compare function for sort.
sub compare_indirectly {
   # print "dbg: compare_indirectly, a=$a, b=$b, ofs(a)=" . $offsets[$a] . ", ofs(b)=" . $offsets[$b] . "\n";
   $offsets[$a] <=> $offsets[$b];
}


# Reorder an array according to an array of indices.  Example:
#   @indices = (0, 4, 1, 3, 2);
#   @values = (100, 101, 102, 103, 104 );
#   @result = &index_by( $#indices, @indices, @values );
# and @result is:
#   (100, 104, 101, 103, 102)
# If Perl were better designed, instead of calling this function you could just do this:
#   @result = $values[ @indices ];
# (What actually happens in that case is that @indices is evaluated in a scalar
# context, and returns the number of elements in the array, which is 5.  So
# $values[@indices] = $values[5] = undefined.)
sub indexby {
   local( $numindices ) = 1 + shift @_;
   local( @indices ) = splice( @_, 0, $numindices );
   local( @values ) = @_;
   local( @result, $index );
   @result = ();
   foreach $index (@indices) {
      push( @result, $values[$index] );
   }
   return @result;
}


# test it:
# @indices = (0, 4, 1, 3, 2);
# @values = (100, 101, 102, 103, 104 );
# @result = &indexby( $#indices, @indices, @values );  # @result = $values[ @indices ];
# print "dbg: (" . join(',',@result) . ")\n\n";  # should be (100,104,101,103,102)


# combine adjacent intervals with the same status (makes the log file smaller)
sub consolidate_adjacent {
   local( $i );
   if ($#offsets != ($numranges-1)) {
      printf "ERR: numranges=%d   <>  (\$\#offsets+1)=%d\n", $numranges, 1+$#offsets;
   }
   for ( $i=1; $i<=$#offsets; ) {
      if ($offsets[$i] == ($offsets[$i-1]+$lengths[$i-1])) {
         $lengths[$i-1] += $lengths[$i];
         splice( @offsets, $i, 1 );
         splice( @lengths, $i, 1 );
         $numranges--;
      } else {
         $i++;
      }
   }
}


# '-i'
if ($report_important) {
   # For now, this is just the MFT.  We look for something like this:

   # Master File Table ($Mft)
   #     $STANDARD_INFORMATION (resident)
   #     $FILE_NAME (resident)
   #     $DATA (nonresident)
   #         logical sectors 6539480-6564583 (0x63c8d8-0x642ae7)
   #         logical sectors 16196368-16355007 (0xf72310-0xf98ebf)
   #         logical sectors 47387448-47388159 (0x2d31338-0x2d315ff)
   #     $BITMAP (nonresident)
   #         logical sectors 6539472-6539479 (0x63c8d0-0x63c8d7)
   #         logical sectors 6216824-6216831 (0x5edc78-0x5edc7f)
   #         logical sectors 14152912-14152919 (0xd7f4d0-0xd7f4d7)

   $found_mft = $numranges = 0;
   seek( FROM_NFI, 0, 0 );  # rewind input file, in case they also specified '-d'
   while (<FROM_NFI>) {
      $_ =~ s/\s*$//g;  # remove trailing whitespace
      if ($_ =~ /^Master File Table \(\$Mft\)/) {
         $found_mft = 1;
         last;
      }
   }
   if ($found_mft) {
      while (<FROM_NFI>) {
         $_ =~ s/\s*$//g;  # remove trailing whitespace
         if ('' eq $_) {
            last;
         }
         if ($_ =~ /^\s*logical sectors\s+(\d+)\-(\d+)\s/) {
            $offsets[$numranges] = 0+$1;
            $lengths[$numranges] = 1 + $2 - $1;
            $numranges++;
         }
      }
      $count_sectors = 0;
      for ($i=0; $i<$numranges; $i++) {
         $len = $lengths[$i];
         $count_sectors += $len;
      }
      print "\$Mft contains $count_sectors sectors in $numranges intervals.\n";

      # Sort the offset/length pairs:
      @indices = &range( 0, $#offsets );
      @indices = sort compare_indirectly @indices;
      @offsets = &indexby( $#indices, @indices, @offsets );
      @lengths = &indexby( $#indices, @indices, @lengths );

      # Generate a ddrescue-style log file (sector list file):
      open( LOG_OUTPUT, ">important.log" ) || die "ERROR: could not write 'important.log', $!\n";
      print LOG_OUTPUT "# Rescue Logfile. Created by GNU ddrescue version 1.1-burtonhack4 (\$Mft)\n" .
            "#      pos        size  status\n";
      $prev_ofs_plus_len = -$partition_offset;
      # first, output the ranges spanning the $Mft:
      for ($i = 0; $i < $numranges; $i++) {
         $ofs = $offsets[$i];
         $len = $lengths[$i];
         if ($ofs < $prev_ofs_plus_len) {
            print "ERROR: overlapping regions:\n  " . &displayrange($ofs+$partition_offset,$len,' ') . "\n";;
         } elsif ($ofs > $prev_ofs_plus_len) {
            $gaplen = $ofs - $prev_ofs_plus_len;
            print LOG_OUTPUT &displayrange($prev_ofs_plus_len+$partition_offset,$gaplen,'+') . "\n";
         }
         print LOG_OUTPUT &displayrange($ofs+$partition_offset,$len,'-') . "\n";
         $prev_ofs_plus_len = $ofs + $len;
      }
      # then output the range of sectors after the $Mft:
      if ($disk_size > 0) {
         $gaplen = $disk_size - ($prev_ofs_plus_len+$partition_offset);
         if ($gaplen > 0) {
            print LOG_OUTPUT &displayrange($prev_ofs_plus_len+$partition_offset,$gaplen,'+') . "\n";
         }
      }
      close LOG_OUTPUT;
      print "Created 'important.log'\n";

      # Generate a ddrescue-style log file (sector list file) for everything else:
      open( LOG_OUTPUT, ">lessimportant.log" ) || die "ERROR: could not write 'lessimportant.log', $!\n";
      print LOG_OUTPUT "# Rescue Logfile. Created by GNU ddrescue version 1.1-burtonhack4 (everything except \$Mft)\n" .
            "#      pos        size  status\n";
      # first, output the 0..firstsector-1 range:
      $prev_ofs_plus_len = -$partition_offset;
      # first, output the ranges spanning the $Mft:
      for ($i = 0; $i < $numranges; $i++) {
         $ofs = $offsets[$i];
         $len = $lengths[$i];
         if ($ofs > $prev_ofs_plus_len) {
            $gaplen = $ofs - $prev_ofs_plus_len;
            print LOG_OUTPUT &displayrange($prev_ofs_plus_len+$partition_offset,$gaplen,'-') . "\n";
         }
         print LOG_OUTPUT &displayrange($ofs+$partition_offset,$len,'+') . "\n";
         $prev_ofs_plus_len = $ofs + $len;
      }
      # then output the range of sectors after the $Mft:
      if ($disk_size > 0) {
         $gaplen = $disk_size - ($prev_ofs_plus_len+$partition_offset);
         if ($gaplen > 0) {
            print LOG_OUTPUT &displayrange($prev_ofs_plus_len+$partition_offset,$gaplen,'-') . "\n";
         }
      }
      close LOG_OUTPUT;
      print "Created 'lessimportant.log'\n";

   } else {
      print "\$Mft was not found in '$from_nfi'.\n" .
            "That's probably good news, meaning that \$Mft wasn't damaged.\n";
   }
}


# '-u'
if ($report_unimportant) {
   # For now, this just lists the freespace.  We're looking for lines like this:

   # ***Logical sector 967280 (0xec270) on drive H is not in any file.
   $numranges = 0;
   undef @offsets;
   undef @lengths;

   if ($verbose) {
      print "-u report:  analyzing freespace sectors...\n";
   }

   seek( FROM_NFI, 0, 0 );  # rewind input file, in case they also specified '-i' or '-d'
   while (<FROM_NFI>) {
      if ($_ =~ /^\*\*\*Logical sector\s(\d+)\s.*is not in any file/) {
         $sect = 0+$1;
         # We know that if any sector in a cluster is unused, then all the
         # sectors in the cluster are unused.  So figure out the sector range
         # for the whole cluster:
         $firstsect = $cluster_size * int( $sect / $cluster_size );
         $offsets[$numranges] = $firstsect;
         $lengths[$numranges] = $cluster_size;
         $numranges++;
         if ($verbose) {
            $lastsect = $firstsect + $cluster_size - 1;
            print "dbg: sector $sect is in the cluster with sectors $firstsect..$lastsect\n";
         }
      }
   }
   if (0 == $numranges) {
      print "Sorry, no unused sectors were named in '$from_nfi'.\n";
   } else {
      # Sort the offset/length pairs:
      @indices = &range( 0, $#offsets );
      @indices = sort compare_indirectly @indices;
      @offsets = &indexby( $#indices, @indices, @offsets );
      @lengths = &indexby( $#indices, @indices, @lengths );

      # remove duplicates:
      for ($i = 1; $i <= $#offsets; $i++) {
         if ($offsets[$i-1] == $offsets[$i]) {
            splice( @offsets, $i, 1 );
            splice( @lengths, $i, 1 );
            $numranges--;
         }
      }
      print "$numranges unused clusters were mentioned in '$from_nfi'.\n";

      # merge adjacent clusters:
      &consolidate_adjacent;
      print "$numranges intervals remain after consolidating adjacent clusters.\n";

      # Generate a ddrescue-style log file (sector list file):
      open( LOG_OUTPUT, ">unimportant.log" ) || die "ERROR: could not write 'unimportant.log', $!\n";
      print LOG_OUTPUT "# Rescue Logfile. Created by GNU ddrescue version 1.1-burtonhack4 (freespace)\n" .
            "#      pos        size  status\n";
      $prev_ofs_plus_len = -$partition_offset;
      for ($i = 0; $i < $numranges; $i++) {
         $ofs = $offsets[$i];
         $len = $lengths[$i];
         if ($ofs < $prev_ofs_plus_len) {
            print "ERR: overlapping regions:\n  " . &displayrange($ofs+$partition_offset,$len,'+') . "  (should be impossible!)\n";;
         } elsif ($ofs > $prev_ofs_plus_len) {
            $gaplen = $ofs - $prev_ofs_plus_len;
            print LOG_OUTPUT &displayrange($prev_ofs_plus_len+$partition_offset,$gaplen,'/') . "\n";
         }
         print LOG_OUTPUT &displayrange($ofs+$partition_offset,$len,'+') . "\n";
         $prev_ofs_plus_len = $ofs + $len;
      }
      if ($disk_size > 0) {
         $gaplen = $disk_size - ($prev_ofs_plus_len+$partition_offset);
         if ($gaplen > 0) {
            print LOG_OUTPUT &displayrange($prev_ofs_plus_len+$partition_offset,$gaplen,'/') . "\n";
         }
      }
      close LOG_OUTPUT;
      print "Created 'unimportant.log'\n";
   }
}
close FROM_NFI;

exit(0);

__END__
